#include    "linkage.h"
#include    "head.inc"
#include	"Interrupts.h"

VECT_TBL	= 0x200000
RDR0		= 0xffffb5
SSR0		= 0xffffb4

        .h8300h
        .file "isr.S"
        .section .text
        .align 2
saveRegisters:
        mov.l   er0, @SYMBOL_NAME(registers)	; save R0
        mov.l   #SYMBOL_NAME(registers), er0
        mov.l   er1, @(0x04, er0)		; save R1
        mov.l   er2, @(0x08, er0)		; save R2
        mov.l   er3, @(0x0c, er0)		; save R3
        mov.l   er4, @(0x10, er0)		; save R4
        mov.l   er5, @(0x14, er0)		; save R5
        mov.l   er6, @(0x18, er0)		; save R6

        mov.l   #0x0, er1				; zero the CCR location
        mov.l   er1, @(0x20, er0)
        mov.w   @er7, r1				; get off the stack(CCR)
        mov.b   r1h, @(0x23, er0)		; save CCR

        mov.l   @er7, er1				; get off the stack(CCR & PC)
        and.l   #0xFFFFFF, er1			; 24-bit addr
        mov.l   er1, @(0x24, er0)		; save PC
        
        adds.l	#4, er7
        mov.l   er7, @(0x1c, er0)		; save R7

        mov.l   @SYMBOL_NAME(stub_sp), er7		; load __stub_stack
        mov.l   @SYMBOL_NAME(exception), er0
        jsr     SYMBOL_NAME(handle_exception)
        mov.l   er7, @SYMBOL_NAME(stub_sp)		; save __stub_stack

;----------------------------------------------------------------------------
restoreRegisters:
        mov.l   #SYMBOL_NAME(registers), er0
        mov.l   @(0x1c, er0), er7				; restore R7
		mov.l	@(0x24, er0), er1				; PC
		sub.w	r2, r2
        mov.b   @(0x23, er0), r2h				; CCR
        or.w	r2, e1
		push.l	er1
        mov.l   @(0x18, er0), er6				; restore R6
        mov.l   @(0x14, er0), er5				; restore R5
        mov.l   @(0x10, er0), er4				; restore R4
        mov.l   @(0x0c, er0), er3				; restore R3
        mov.l   @(0x08, er0), er2				; restore R2
        mov.l   @(0x04, er0), er1				; restore R1
        mov.b	#0x00, r0l
        mov.b	r0l, @SYMBOL_NAME(in_nmi)		; in_nmi = 0        
        mov.l   @SYMBOL_NAME(registers), er0	; restore R0
        rte										; restore CCR & PC

;----------------------------------------------------------------------------
ENTRY(INT_NMI)				/* NMI */
        push.l	er0
        mov.l	#IntID_NMI, er0
        mov.l	er0, @SYMBOL_NAME(exception)
        mov.b	@er0, r0l
        beq		NEXT
        pop.l	er0
        rte
NEXT:   mov.b	#0x01, r0l
        mov.b	r0l, @SYMBOL_NAME(in_nmi)
        pop.l	er0
        bra		saveRegisters:16

;----------------------------------------------------------------------------
ENTRY(INT_TRAP0)			/* BREAK */
        push.l	er0
        mov.l	#IntID_TRAP0, er0
        mov.l	er0, @SYMBOL_NAME(exception)
        pop.l	er0
        bra		saveRegisters:16

;----------------------------------------------------------------------------
ENTRY(INT_TRAP1)			/* BIOS CALL */
        push.l	er0
        mov.l	#IntID_TRAP1, er0
        mov.l	er0, @SYMBOL_NAME(exception)
        pop.l	er0
        bra		saveRegisters:16

;----------------------------------------------------------------------------
ENTRY(INT_TRAP2)
        push.l	er0
        mov.l	#IntID_TRAP2, er0
        mov.l	er0, @SYMBOL_NAME(exception)
        pop.l	er0
        bra		saveRegisters:16

;----------------------------------------------------------------------------
ENTRY(INT_TRAP3)			/* BRAEK */
        push.l	er0
        mov.l	#IntID_TRAP3, er0
        mov.l	er0, @SYMBOL_NAME(exception)
        pop.l	er0
        bra		saveRegisters:16

;----------------------------------------------------------------------------
ENTRY(INT_IRQ0)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IRQ0, er0
jmp_vect:
        mov.l	er0, @SYMBOL_NAME(exception)
        push.l	er1
        shll.l	er0
        shll.l	er0
        add.l	#VECT_TBL, er0
        mov.l	@er0, er1
        mov.w	@(0x0a, er7), r0
        mov.b	#0x00, r0l
        or.w	r0, e1
        mov.l	er1, @(0x08, er7)
        pop.l	er1
        pop.l	er0
        rte

;----------------------------------------------------------------------------
ENTRY(INT_IRQ1)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IRQ1, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IRQ2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IRQ2, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IRQ3)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IRQ3, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IRQ4)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IRQ4, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IRQ5)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IRQ5, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_WOVI)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_WOVI, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_CMI)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_CMI, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_ADI)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_ADI, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IMIA0)		/* 16 BIT TIMER 0 */
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IMIA0, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IMIB0)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IMIB0, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_OVI0)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_OVI0, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IMIA1)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IMIA1, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IMIB1)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IMIB1, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_OVI1)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_OVI1, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IMIA2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IMIA2, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_IMIB2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_IMIB2, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_OVI2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_OVI2, er0
        bra		jmp_vect:16
        
;----------------------------------------------------------------------------
ENTRY(INT_CMIA0)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_CMIA0, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_CMIB0)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_CMIB0, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_CMIA1B1)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_CMIA1B1, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_TOVI01)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_TOVI01, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_CMIA2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_CMIA2, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_CMIB2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_CMIB2, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_CMIA3B3)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_CMIA3B3, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_TOVI23)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_TOVI23, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_DEND0A)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_DEND0A, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_DEND0B)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_DEND0B, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_DEND1A)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_DEND1A, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_DEND1B)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_DEND1B, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_ERI0)
        push.l	er0
        mov.l	#IntID_ERI0, er0
        bra		handle_break

;----------------------------------------------------------------------------
ENTRY(INT_RXI0)			/* SCI0 Rx */
        push.l	er0
        mov.l	#IntID_RXI0, er0
handle_break:
        mov.l	er0, @SYMBOL_NAME(exception)
        mov.b	@RDR0, r0l
        mov.b	@SSR0, r0h
        and.b	#0x80, r0h				; RDRF & Error clear
        mov.b	r0h, @SSR0
        cmp.b	#0x03, r0l
        beq		sci_break
        pop.l	er0
        rte
sci_break:        
        pop.l	er0
        bra		saveRegisters:16

;----------------------------------------------------------------------------
ENTRY(INT_TXI0)
        push.l	er0
        mov.l	#IntID_TXI0, er0
        mov.l	er0, @SYMBOL_NAME(exception)
        pop.l	er0
        rte

;----------------------------------------------------------------------------
ENTRY(INT_TEI0)
        push.l	er0
        mov.l	#IntID_TEI0, er0
        mov.l	er0, @SYMBOL_NAME(exception)
        pop.l	er0
        rte

;----------------------------------------------------------------------------
ENTRY(INT_ERI1)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_ERI1, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_RXI1)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_RXI1, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_TXI1)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_TXI1, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_TEI1)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_TEI1, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_ERI2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_ERI2, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_RXI2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_RXI2, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_TXI2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_TXI2, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_TEI2)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_TEI2, er0
        bra		jmp_vect:16

;----------------------------------------------------------------------------
ENTRY(INT_Reserved)
        stc.w	ccr, @-er7
        subs	#2, er7
        push.l	er0
        mov.l	#IntID_TEI2, er0
        bra		jmp_vect:16

